"""
Rollet settings and its methods
"""

__all__ = ["ABSTRACT"]

from rollet.extractor import *

import os 
dir_path = os.path.dirname(os.path.realpath(__file__))


# CSS rules selectors for each field
ABSTRACT = (
    *[f':is(div,section)[{attr}*="{value}" i]' for attr in ('class', 'id')
    for value in ('-abstract', 'abstract')],
    'meta[name="dc.description" i]',
    'meta[property="og:description" i]',
    *[f'meta[name="{attr}" i]'
    for attr in ('description', 'twitter:description', 'citation_abstract')],
    *[f'div[{attr}*="article" i][{attr}*="body" i]'
    for attr in ('class', 'itemprop')],
    'div[class*=content i]',
    'script[type*="json" i]',
    *[f':is(div,section)[{attr}*="abs" i]' for attr in ('class', 'id')],
    *[f'div[class="{attr}" i]' for attr in ('desc', 'new-txt', 'contenuto')]
)

TITLE = (
    'meta[property="og:title" i]',
    *[f'meta[name="{attr}" i]'
    for attr in ('citation_title', 'dc.title', 'twitter:title', 'title')],
    'script[type*="json" i]',
    *[f':is(h1,h2,h3,div,span)[{attr}*=title i]'
    for attr in ('class', 'id', 'itemprop')],
    *[f':is(h1,h2,h3)[itemprop*=head i]'],
    'title'
)

# Domains types
types = {
    'article': [
        'agronews', 'ansa', 'berkeley', 'citrusindustry', 'europa', 'europafm',
        'europalibera', 'europapress', 'europatat', 'floraldaily',
        'freshfruitportal', 'freshplaza', 'groentennieuws', 'hortidaily',
        'hortweek', 'huffingtonpost', 'ica', 'mon-viti', 'nieuweoogst', 'nvwa',
        'oliveoiltimes', 'phytoma', 'phytoma-ldv', 'teatronaturale',
        'terrenature', 'theguardian', 'usda', 'vniikr'
    ],
    'paper': [
        'apsnet', 'biorxiv', 'chinapubmed', 'europepmc', 'mdpi', 'nature',
        'nih', 'nl', 'researchgate', 'sciencedirect', 'springer',
        'srpingeropen', 'wiley'
    ]
}
TYPES = dict(sum([[(i,k) for i in v] for k,v in types.items()], []))

# Types defaults keywords arguments for fetch
FETCH_KWARGS = {
    'default': {
        'abstract__which': 'max',
        'abstract__script_keys': [],
        'lang__attr': 'lang',
        'title__which': 'first',
        'title__script_keys': []
    },'article': {
        'abstract__which': 'max',
        'abstract__script_keys': ['abstract', 'summary'],
        'lang__attr': 'lang',
        'title__which': 'first',
        'title__script_keys': ['title', 'heading']
    },
    'paper': {
        'abstract__which': 'first',
        'abstract__script_keys': ['abstract', 'summary'],
        'lang__attr': 'lang',
        'title__which': 'first',
        'title__script_keys': ['title', 'heading', 'name']
    },
}

# Blacklisted domains
with open(os.path.join(dir_path, 'blacklist'), errors = 'ignore') as f:
    BLACKLIST = list(map(lambda x:x.strip(), f.readlines()))

# Custom domains extractors
# {domain: [(schema1, Extractor1), (schema2, Extractor2)]}
EXTRACTORS = {
    'default': [(r'.*', BaseExtractor),],
    'mdpi': [(r'.*', MdpiExtractor),],
    'nih': [(r'.*', NihExtractor),],
    'sciencedirect': [(r'.*', ScienceDirectExtractor),],
}

# Methods

def add_types(dict_types):
    """
    Add new type domains
    dict_types: dict, {type: [domain1, domain2, ...],}
    :return: None
    """
    global TYPES
    news = dict(sum([[(i,k) for i in v] for k,v in dict_types.items()], []))
    TYPES = {**TYPES, **news}


def add_fetch_kwargs(dict_types):
    """
    Add new fetch kwargs for type domains
    dict_types: dict, {type: {kwargs1:value1, kwargs2:value2,},}
    :return: None
    """
    global FETCH_KWARGS
    news = FETCH_KWARGS.copy()
    for d_type, key_args in dict_types.items():
        news[d_type] = {**news.get(d_type, news['default']), **key_args}
    FETCH_KWARGS = news.copy()



import re

def rfinditem(obj, key):
    """
    Find key recursively in dict
    obj: Dict
    key: str
    :return: None or dict item
    """
    if key in obj: return obj[key]
    for k, v in obj.items():
        if isinstance(v, dict):
            item = rfinditem(v, key)
            if item is not None:
                return item


def lookup(term, zipped):
    """
    Find in zipped term's value if match pattern
    term: str
    zipped: list of tuples
    :return: None or value
    """
    for pattern, value in zipped:
        if re.search(pattern, term):
            return value
    return None


"""
Rollet runtime console script
"""
import argparse
import pandas as pd
from time import sleep
from tqdm import tqdm
from datetime import datetime

from rollet import get_content


def extract_col(link, fields, timesleep = 0):
    try:
        content = get_content(link)
    except:
        content = {
            'title': 'failed',
            'date': datetime.now().strftime("%m/%d/%Y, %H:%M:%S")}
    finally:
        content = [content.get(k) for k in fields]
    return pd.Series(content)


def extract_csv(
    path, link_col,
    fields, **kwargs
):
    path_out = f'{path[:-4]}-output.csv' if not kwargs.get('outfile') else kwargs['outfile']
    start = kwargs.get('start', 0)
    size = None if kwargs.get('size', -1) < 0 else kwargs.get('size', 100)
    timesleep = kwargs.get('timesleep', 0)
    df = pd.read_csv(path, nrows = size, skiprows = lambda x: x>1 and x<=start)
    tqdm.pandas()
    df[fields] = df[link_col].progress_apply(extract_col, args = (fields, timesleep,))
    df.to_csv(path_out, index = False)


def rollet_extract():
    parser = argparse.ArgumentParser(prog = 'rollet')
    parser.add_argument(
        'extract',
        choices = ['extract-txt', 'extract-csv', 'extract-json'],
        help = 'Choose file type option extraction')
    parser.add_argument('path', help = 'file path')
    parser.add_argument(
        '-o', '--outfile', nargs = '?',
        help = 'output file path')
    parser.add_argument(
        '-l', '--link', nargs = '?',
        help = 'link field if csv or json')
    parser.add_argument(
        '-f', '--fields', nargs = '?',
        default = "status,title,abstract,lang,content_type,date",
        help = 'fields to keep separated by comma')
    parser.add_argument(
        '--start', nargs = '?', default = 0, type = int,
        help = 'number of rows to skip')
    parser.add_argument(
        '--size', nargs = '?', default = -1, type = int,
        help = 'max number of rows to keep')
    parser.add_argument(
        '-t', '--timesleep', nargs = '?', default = 0,
        help = 'sleep time in seconds between two pulling')

    args = parser.parse_args()
    args.timesleep = float(args.timesleep)
    args.fields = args.fields.split(',')
    if (args.extract != 'extract-txt') and not args.link:
        raise ValueError(
            """The --links field name must be filled in
            if the data is not a text column extract-txt""")
    
    if args.extract == "extract-csv":
        extract_csv(
            args.path, args.link, args.fields,
            start = args.start,
            size = args.size,
            outfile = args.outfile,
            timesleep = args.timesleep
        )
    
    else:
        raise NotImplementedError('other extraction format is not supported yet')



__all__ = [
    "BaseExtractor",
    "MdpiExtractor",
    "NihExtractor",
    "ScienceDirectExtractor",
]

import re
from requests import get
from requests.exceptions import InvalidURL
from requests.utils import default_headers
from bs4 import BeautifulSoup as B
from tldextract import extract
from html import unescape
from json import loads
from pprint import pprint as cat, pformat as pcat
from datetime import datetime


from rollet import settings
from rollet.utils import (
    rfinditem
)


class BaseExtractor:
    def __init__(self, url, **kwargs):
        """
        Create BaseExtractor instance
        url: string
        timeout: float, request timeout. Default 1 sec.
        abstract_**kwargs: **kwargs for abstract fetch
        title_**kwargs: **kwargs for title fetch
        """
        
        self.url = url
        self.domain = extract(url).domain
        self.date = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        self._scrap(**kwargs)
        self._init_kwargs(**kwargs)
    
    def _scrap(self, **kwargs):
        timeout = kwargs.pop('timeout', 1)
        headers = kwargs.pop('headers', None)
        with get(self.url, timeout = timeout, headers = headers) as response:
            self._status = response.status_code
            self._header = response.headers
            self._content_type()
            if self.content_type == 'html':
                self._page = B(response.text, 'html.parser')
            else:
                self._page = B()
    
    def _init_kwargs(self, **kwargs):
        self._abstract_kwargs, self._title_kwargs, self._lang_kwargs = {}, {}, {}
        for k,v in kwargs.items():
            field, key = k.split('__')
            if field == 'abstract': self._abstract_kwargs.update({key:v})
            elif field == 'title': self._title_kwargs.update({key:v})
            elif field == 'lang': self._lang_kwargs.update({key:v})
    

    def _content_type(self):
        if self.url[-3:] == 'pdf': content = 'pdf'
        else:
            charset = self._header.get('Content-Type', '')
            content = re.findall('(html|pdf|json|xml)', charset)
            content = content[0] if len(content) else 'html'
        self.content_type = content


    @staticmethod
    def _get_content(tag, **kwargs):
        """
        Get content from element Tag
        tag: bs4.element.Tag
        script_keys: List, of keys if tag is script
        attr: str, key attribute if tag is neither a meta nor a script
        :return: content
        """

        if tag.name == 'meta':
            content = tag.attrs.get('content', [])
        elif tag.name == 'script':
            keys = kwargs.get('script_keys', [])
            try:
                serie = loads(tag.content[0])
            except:
                content = list()
            else:
                content = [rfinditem(serie, key) for key in keys]
        elif kwargs.get('attr'):
            content = tag.attrs.get(kwargs.get('attr'))
        else:
            content = tag.text.strip().replace('\n', '')

        content = content[0] if isinstance(content, list) and len(content) else content
        content = unescape(content) if isinstance(content, str) else content
        return content
    

    def __repr__(self):
        string = "Title: {}\nFrom: {}\nFetched at: {}\nStatus: {}\nType: {}\n{} Abstract {}\n{}"
        return string.format(
            self.title, self.url, self.date,
            self._status, self.content_type,
            '-'*5, '-'*5, pcat(self.abstract), 
        )


    def fetch(self, selectors, which='first', **kwargs):
        content = list()
        arg_w = ('first', 'min', 'max')
        if which not in arg_w:
            raise ValueError(f'which should be one of {arg_w}')
        for s in selectors:
            contents_ = list()
            tags = self._page.select(s)
            if len(tags): contents_ = [self._get_content(tag, **kwargs) for tag in tags]
            if len(contents_) and which == 'first': 
                content = [contents_[0]]
                break
            else:
                try: content += ['. '.join(set(contents_))]
                except: pass
        content = content if len(content) else [None]
        if which == 'max': content = max(content, key=lambda x: len(str(x)))
        else: content = min(content, key=lambda x: len(str(x)))
        return content
    
    @property
    def title(self):
        title = None
        if self.content_type == 'html':
            title = self.fetch(settings.TITLE, **self._title_kwargs)
        return title
    
    @property
    def abstract(self):
        abstract = None
        if self.content_type == 'html':
            abstract = self.fetch(settings.ABSTRACT, **self._abstract_kwargs)
        return abstract
    
    @property
    def lang(self):
        lang = None
        if self.content_type == 'html':
            lang = self._page.html.get('lang', None)
        return lang
    
    
    def to_dict(self):
        return {
            'url': self.url,
            'status': self._status,
            'title': self.title,
            'abstract': self.abstract,
            'lang': self.lang,
            'content_type': self.content_type,
            'date': self.date
        }
    
    def to_list(self, *args):
        if len(args): listed = [getattr(self, arg, None) for arg in args]
        else: listed = list(self.to_dict().values())
        return listed


class MdpiExtractor(BaseExtractor):
    """
    Extractor for base mdpi.com domain
    """

    @property
    def abstract(self):
        abstract = self._page.find(class_='art-abstract').text.strip()
        abstract = super().abstract if not abstract else abstract
        return abstract

    @property
    def title(self):
        title = self._page.find(class_='title', attrs={
            'itemprop': 'name'
        }).text.strip()
        title = super().title if not title else title
        return title
    

class NihExtractor(BaseExtractor):
    """
    Extractor for base nih.gov domain
    """
    
    @property
    def abstract(self):
        abstract = self._page.find(id='enc-abstract').text.strip()
        abstract = super().abstract if not abstract else abstract
        return abstract

    @property
    def title(self):
        title = self._page.find('h1', class_='heading-title').text.strip()
        title = super().title if not title else title
        return title


class ScienceDirectExtractor(BaseExtractor):
    """
    Extractor for base sciencedirect.com domain
    """
    def _scrap(self, **kwargs):
        headers = default_headers()
        headers.update({
            'User-Agent':
            'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0',
        })
        super()._scrap(headers=headers, **kwargs)

    @property
    def abstract(self):
        abstract = self._page.find(id='abstracts',
                                   class_='Abstracts').text.strip()
        abstract = super().abstract if not abstract else abstract
        return abstract

    @property
    def title(self):
        title = self._page.find('span', class_='title-text').text.strip()
        title = super().title if not title else title
        return title


"""
Rollet functions
"""

__all__ = [
    "get_content"
]

from datetime import datetime
from tldextract import extract
from requests.exceptions import ConnectTimeout

from rollet import settings
from rollet.utils import lookup
from rollet.extractor import (
    BaseExtractor
)

def get_content(url):
    """
    Pull content from url to formated data
    url: string url
    :return: dict of content
    """
    content = {
        'url': url,
        'status': None,
        'title': 'blacklisted',
        'abstract': None,
        'lang': None,
        'content_type': None,
        'date': datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
    }
    domain = extract(url).domain
    if domain in settings.BLACKLIST: return content
    f_kwargs = settings.FETCH_KWARGS[settings.TYPES.get(domain, 'default')]
    extractor = lookup(
        url,
        settings.EXTRACTORS.get(domain, settings.EXTRACTORS['default']))
    extractor = extractor if extractor else BaseExtractor
    try:
        content = extractor(url, **f_kwargs).to_dict()
    except ConnectTimeout:
        content['title'] = "Timeout error"
    except Exception as e:
        content['title'] = str(e)
    return content


from setuptools import setup, find_packages
import codecs
import os
import rollet

here = os.path.abspath(os.path.dirname(__file__))

long_description = open(os.path.join(here, "README.md"), encoding="utf8").read()

with open(os.path.join(here, "requirements.txt")) as fq:
    requirements = list(map(lambda x:x.strip(), fq.readlines()))

VERSION = rollet.__version__
DESCRIPTION = 'Collect data from various sources'

setup(
    name = "rollet",
    version = VERSION,
    author = "Opscidia (Tech)",
    author_email = "tech@opscidia.com",
    maintainer = "Loïc Rakotoson",
    maintainer_email = "loic.rakotoson@opscidia.com",
    description = DESCRIPTION,
    long_description_content_type = "text/markdown",
    long_description = long_description,
    license_files = ('LICENSE',),
    packages = find_packages(),
    include_package_data = True,
    install_requires = requirements,
    keywords = ['fetch', 'pull', 'extract', 'scrap'],
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires = '>=3.7',
    entry_points = {
        'console_scripts': [
            'rollet = rollet.runtime:rollet_extract'
        ]
    },
)